/**
 * Copyright 2010 Google Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.google.diffable.listeners;

import java.io.File;
import java.io.FileInputStream;
import java.util.ArrayList;
import java.util.Properties;

import javax.servlet.ServletContext;
import javax.servlet.ServletContextEvent;
import javax.servlet.ServletContextListener;

import org.apache.log4j.Logger;

import com.google.diffable.config.BaseModule;
import com.google.diffable.config.MessageProvider;
import com.google.diffable.data.ResourceManager;
import com.google.diffable.exceptions.ResourceManagerException;
import com.google.diffable.exceptions.StackTracePrinter;
import com.google.diffable.tags.DiffableResourceTag;
import com.google.inject.Guice;
import com.google.inject.Inject;
import com.google.inject.Injector;

public class DiffableListener implements ServletContextListener {
	@Inject(optional=true)
	private StackTracePrinter printer;
	
	@Inject(optional=true)
	private ResourceManager mgr = null;
	
	@Inject(optional=true)
	private Logger logger = Logger.getLogger(DiffableListener.class);
	
	@Inject(optional=true)
	private MessageProvider provider;
	
	private BaseModule baseModule = null;
	
	private Injector inj = null;
	
	private ResourceMonitor monitor;
	
	public void contextDestroyed(ServletContextEvent ctxEvent) {
		this.monitor.stopProcessing();
		try {
			this.monitor.join();
		} catch (InterruptedException exc) {
			printer.print(exc);
		}
	}
	
	public void contextInitialized(ServletContextEvent ctxEvent) {
		ServletContext ctx = ctxEvent.getServletContext();
	    baseModule = new BaseModule();
	    
		// Initialize Guice provided properties.
		inj = initializeInjectedProperties(ctx, baseModule);
		
		// Save the injector for later use by Diffable.
		ctx.setAttribute("DiffableGuiceInjector", inj);
		
		// Initialize the resource manager. For testing, the DiffableListener
		// is retrieved via Guice with a mock ResourceManager injected already,
		// which is why this conditional check is performed.
		if (mgr == null) {
			mgr = inj.getInstance(ResourceManager.class);
		}
		mgr.setServletContext(ctx);
		
		// By retrieving the message provider via the injector, the message
		// bundles path can be overridden.
		provider = inj.getInstance(MessageProvider.class);
		printer = inj.getInstance(StackTracePrinter.class);
		
		// Initialize the resource manager and store it in the servlet context.
		try {
			ctx.setAttribute("ResourceManager", mgr.initialize());
		} catch (ResourceManagerException exc) {
			provider.error(logger, "resourcestore.problem");
			printer.print(exc);
		}
		
		// Retrieve the directories containing resources to be managed and
		// crawl them for resources, adding them to the resource manager as
		// they are found.  If the path begins with a slash, it will be
		// interpreted as absolute. Otherwise, it will be interpreted as
		// relative to the web app directory. If multiple folders are specified,
		// they should be separated by commas.
		String currentPath = ctx.getRealPath("");
		String resourceFolders = ctx.getInitParameter("ResourceFolders");
		ArrayList<File> foundFolders = new ArrayList<File>();
		if (resourceFolders == null) {
			String error =
				provider.getMessage(
					"ErrorMessages", "resourcefolders.none");
			logger.error(error);
		} else {
			String[] folders = resourceFolders.split(",");
			for (String path : folders) {
				File folder = path.startsWith("/") ?
					new File(path) :
					new File(currentPath + File.separator + path);
				if (!folder.exists()) {
					provider.error(logger, "resourcefolders.nonexistent",
								   folder.getAbsolutePath());
				} else if (!folder.isDirectory()) {
					provider.error(logger, "resourcefolders.notfolder",
								   folder.getAbsolutePath());
				} else {
					foundFolders.add(folder);
				}
			}
			// Set the folders on the DiffableTag so it can locate resources.
			DiffableResourceTag.setFolder(foundFolders);
			
			// Start up the monitoring thread.
			monitor = inj.getInstance(ResourceMonitor.class);
			monitor.setFolderAndManager(foundFolders, mgr);
			monitor.setDaemon(true);
			monitor.start();
		}
	}
	
	/**
	 * Private function for initializing Guice provided member variables. Since
	 * the listener is initialized by the servlet container, this needs to be
	 * performed manually outside of testing.
	 */
	protected Injector initializeInjectedProperties(ServletContext ctx, 
			                                        BaseModule module) {
		// Attempt to load the configuration properties if provided by the
		// web.xml under 'DiffableConfigProperties'.  If the file exists, use
		// it to initialize the module.
		String propFilePath = ctx.getInitParameter("DiffableConfigProperties");
		String currentPath = ctx.getRealPath("");
		File propFile = propFilePath.startsWith("/") ?
						new File(propFilePath) :
				        new File(currentPath + File.separator + propFilePath);
		if (propFilePath != null) {
			try {
				Properties props = new Properties();
				props.load(new FileInputStream(propFile));
				module.setProperties(props);
			} catch (Exception exc) {
				new MessageProvider().error(logger, 
						                    "module.propfile",
						                    propFile.getAbsolutePath());
				new StackTracePrinter().print(exc);
			}
		}
		return Guice.createInjector(module);
	}
	
	/**
	 * Recursively descends into a folder, adding resources to the manager as
	 * they are found.
	 * 
	 * @param directory A directory to search for resources.
	 */
	private void addResourcesInPath(File directory) {
		File[] resources = directory.listFiles();
		for (File resource : resources) {
			if (resource.isFile()) {
				try {
					mgr.putResource(resource);
				} catch (Exception exc) {
					provider.error(logger, "resources.cannotput",
						           resource.getAbsolutePath());
					printer.print(exc);
				}
			} else if (resource.isDirectory()) {
				addResourcesInPath(resource);
			}
		}
	}
}
